;; Copyright (c) 1996-1998 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;
;; Defines <app-text>, a text widget for applications that include
;; mini buffers.
;;
;;
;; Marks and Indexes
;; -----------------
;;
;; A mark denotes the char immediately to the right of the mark
;; location (mark locations lie between characters).
;;
;; All marks move when text is inserted before them.
;;
;; Note: _all_ methods that operate on a range of locations, and that
;; take an end location as an argument, perform their operations
;; _exclusive_ of the end location. (e.g., get-string returns a string
;; up to but not including the end location.)
;;
;; end-mk always points just past the last inserted buffer char.
;; We define end-mk because the official Tk end mark (named 'end in swl)
;; always points one past an extra newline character (yecch!).
;;
;; insert is a predefined mark for the insertion point. The text-widget
;; "cursor" is really the insert mark.
;;


(require "../common/help-text.ss" 'once)


(define-class (<app-text> parent) (<help-text> parent)
  (ivars
   (key-prefix #f)
   (end-mk #f)
   (mini-buffer #f)
   (mini-search #f)
   (search-string #f)
   (search-offset #f)
   (search-alive? #f)
   (search-beg #f)
   (search-end #f)
   (search-msg #f)
   (search-markup (create <markup> with (background-color: 'orange)))
   (search-buffer #f)
   )
  (inherited handle)
  (inheritable mini-buffer key-prefix end-mk handle)
  
  (private
   [do-the-search (forward?)
     (define search-action
       (lambda any
	 (send self set-focus)
	 (hide-mini-search)
	 (set! search-msg (if forward? "Forward search" "Backward search"))
	 (send self display-mini search-msg)
	 (let* ((search-string (send mini-search get-string))
		(buflen (string-length search-buffer))
		(len (string-length search-string))
		(match-ofs
		 ((if forward?
		      (lambda (a b ofs) (substring-match a b ofs))
		      (lambda (a b ofs) (substring-match-rev a b ofs)))
		  search-string search-buffer search-offset))
		)
	   (send self display-mini
		 (string-append search-msg ": " search-string))
	   (if match-ofs
	       (begin
		 (when search-beg
		   (send search-markup remove-markup
			 self search-beg search-end))
		 (set! search-beg
		   (send self add-offset '(0 . 0) match-ofs))
		 (set! search-end
		   (send self add-offset '(0 . 0) (+ match-ofs len)))
		 (send self set-cursor-pos! search-beg)
		 (send search-markup apply-markup self search-beg search-end)
		 (set! search-offset match-ofs)
		 )
	       (begin 
		 (send self display-mini
		       (string-append search-msg " failed; wrapping: "
				      search-string))
		 (set! search-offset (if forward? -1 buflen))
		 (set! search-msg
		   (if forward?
		       "Wrapped forward search"
		       "Wrapped backward search"))
		 )))))
      (if search-alive?
	(search-action)
	(let ((prompt (if forward? "Forward Search:" "Backward Search:")))
	  (send self display-mini prompt)
	  (hide mini-buffer)
	  (send mini-buffer set-width/char! (string-length prompt))
	  (pack mini-buffer (expand: #f) (side: 'left))
	  (pack mini-search (expand: #t) (fill: 'x) (side: 'left))
	  (send mini-search set-enabled! #t)
	  (send mini-search set-focus)
	  (set! search-alive? #t)
	  (let ((ofs (send self index->offset (cursor))))
	    (if forward?
		(set! search-offset ofs)
		(set! search-offset (1- ofs))))
	  (set! search-buffer (send self get-string '(0 . 0) end-mk))
	  (send mini-search set-action! search-action)
	  ))
      ]
   ) ;;; end private

  (protected ;;;             ** ** PROTECTED ** **

   [hide-mini-search ()
      (hide mini-search)
      (pack mini-buffer (expand: #t) (side: 'left) (fill: 'x))]
    
   [cursor () (send self get-cursor-pos)]

   [substring-match (str1 str2 ofs)
;;; (printf "substr  s1=~s  s2=~s  ofs=~s~n" str1 str2 ofs)
      ;;
      ;; returns offset of match (of str1 in str2), or #f if none or bad ofs
      ;;
      (let ((len1 (string-length str1))
            (len2 (string-length str2)))
        (if (zero? len1)
          ofs
          (if (negative? ofs)
            #f
            (let loop ((i ofs))
              (if (fx>= i len2)
                #f
                (let loop2 ((j 0) (k i))
                  (cond
                    [(fx= j len1) i]
                    [(fx>= k len2) #f]
                    [(char-ci=? (string-ref str1 j) (string-ref str2 k))
                     (loop2 (fx1+ j) (fx1+ k))]
                    [else (loop (fx1+ i))])))))))]

   [substring-match-anchored (str1 str2 ofs)
;;; (printf "substr-ma:  s1=~s s2=~s ofs=~s~n" str1 str2 ofs)
      ;;
      ;; returns ofs if match or #f if none or bad ofs
      ;;
      (let ((len1 (string-length str1))
            (len2 (string-length str2)))
        (if (zero? len1)
          ofs
          (if (negative? ofs)
            #f
            (let loop ((j 0) (k ofs))
              (cond
                [(fx= j len1) ofs]
                [(fx>= k len2) #f]
                [(char-ci=? (string-ref str1 j) (string-ref str2 k))
                 (loop (fx1+ j) (fx1+ k))]
                [else #f])))))]

   [substring-match-rev (str1 str2 ofs)
;;; (printf "substr-mr  s1=~s  s2=~s  ofs=~s~n" str1 str2 ofs)
      ;;
      ;; returns ofs of match or #f if none or bad ofs
      ;;
      (let* ((len1 (string-length str1))
             (len1m1 (1- len1))
             (len2 (string-length str2)))
        (cond
          [(negative? ofs) #f] ;;; match not possible
          [(fx>= ofs len2) #f] ;;; invalid offset
          [(zero? len1) ofs] ;;; anything matches
          [else
            (let loop ((i ofs))
              (if (fx< i len1m1)
                #f ;;; insufficient chars avail for matching
                (let loop2 ((j len1m1) (k i))
                  (cond
                    [(fxnegative? j) (fx1+ k)]
                    [(char-ci=? (string-ref str1 j) (string-ref str2 k))
                     (loop2 (fx1- j) (fx1- k))]
                    [else (loop (fx1- i))]))))]))]

   ) ;;; end protected

  (public
   [init (parent)
     (send-base self init parent)
     (set! end-mk (send self floating-mark '(0 . 0)))
    ]

   [init-mini (mb ms)
     (set! mini-buffer mb)
     (set! mini-search ms)
     (send mini-search set-escape-thunk!
	   (lambda () (send self turn-search-off)))
     ]
   
   [get-end-mk () end-mk]

   [offset->index (ofs) (send self add-offset '(0 . 0) ofs)]

   [index->offset (idx)
     (define ballpark
       (lambda ()
	 (let loop ((ofs 1))
	   (if (send self pos<=? idx (send self add-offset '(0 . 0) ofs))
	       ofs
	       (loop (+ ofs ofs))))))
     (if (send self pos=? idx 'end) ;;; all idxs > 'end are pos=? to 'end
	 (1+ (send self index->offset (send self add-offset 'end -1)))
	 (let* ((guess-high (ballpark))
		(guess-low (quotient guess-high 2)))
	   (define bsearch
	     (lambda (low high)
	       (let* ((mid (quotient (+ low high) 2))
		      (mid-idx (send self add-offset '(0 . 0) mid)))
		 (cond
		   [(send self pos=? idx mid-idx) mid]
		   [(send self pos<? idx mid-idx) (bsearch low (1- mid))]
		   [else (bsearch (1+ mid) high)]))))
	   (bsearch guess-low guess-high)))]
   
   [clear-mini ()
      (when mini-buffer
	(send mini-buffer set-enabled! #t)
	(send mini-buffer delete 0 'end)
	(send mini-buffer set-enabled! #f)
	)]

   [get-mini ()
     ;;
     ;; returns mini-buffer string (w/o trailing newline) or #f if none
     ;;
     (if mini-buffer
	 (let* ((buf (send mini-buffer get-string))
		(str (substring buf 0 (string-length buf))))
;;;(fprintf (swl:bug-port) "get-mini str=~s~n" str)
	   (and (positive? (string-length str)) str))
	 #f)]

   [display-mini (str)
     (when mini-buffer
;;;(fprintf (swl:bug-port) "display-mini ~s~n" str)
       (send self clear-mini)
       (send mini-buffer set-enabled! #t)
       (send mini-buffer insert-at 0 str)
       (send mini-buffer set-enabled! #f)
       )]

   [delete (idx)
      (send-base self delete idx)
      (send self turn-search-off)
      ]

   [delete (idx1 idx2)
      (send-base self delete idx1 idx2)
      (send self turn-search-off)
      ]

   [insert-at (idx what)
      (send-base self insert-at idx what)
      (send self turn-search-off)
      ]

   [key-press (key mods)
      (if key-prefix
	(send self key-press-prefix key mods)
	(send self key-press-no-prefix key mods))
      ]

   [key-press-prefix (key mods)
      (void)]

   [key-press-no-prefix (key mods)
      (event-case ((key= key) (modifier= mods))
	[([control #\s]) (send self search-forward)]
	[([control #\r]) (send self search-backward)]
	[([up] [down] [left] [right])
	 (send self turn-search-off)
	 (send-base self key-press key mods)
	 ]
	[else
 	  (unless (number? key) ;;; kludge test for modifier key-only
	      (send self turn-search-off)
	      )
	  (send-base self key-press key mods)
	  ]
	)]
    
   [turn-search-off ()
     (when search-alive?
       (set! search-alive? #f)
       (set! search-buffer #f)
       (hide-mini-search)
       (send self clear-mini)
       (when search-beg
	 (send search-markup remove-markup self search-beg search-end)
	 (set! search-beg #f))
       (send self set-focus))]

   [search-forward ()
      (when search-alive? (set! search-offset (1+ search-offset)))
      (do-the-search #t)]

   [search-backward ()
      (when search-alive? (set! search-offset (1- search-offset)))
      (do-the-search #f)]

   [mouse-press (x y mods)
      (send self turn-search-off)
      (send-base self mouse-press x y mods)]
      
   [action-copy ()
      (critical-section
	(let ((selstr (swl:get-selection)))
	  (when selstr
	    (swl:clear-clipboard)
	    (swl:append-clipboard selstr)
	    )))]

   [action-cut ()
     (critical-section
	(let ((selrange (send self get-selected-range)))
	  (when selrange
	    (let ((selstr (swl:get-selection)))
	      (when selstr
		(swl:clear-clipboard)
		(swl:append-clipboard selstr)
		(send self delete (car selrange) (cdr selrange))
		)))))]

   [action-paste ()
     (critical-section
       (let ((selstr (swl:get-clipboard)))
	 (when selstr
	   (send self insert-at (cursor) selstr))))]

    ) ;;; end public
  )



#!eof



(require "../common/scrollframe.ss" 'once)

(define new-app-text 
  (case-lambda
    [(height width)
     (let* ((top (create <app-toplevel>))
	    (scrolled-frame
              (create <scrollframe> top with
		(default-vscroll: #t) (sticky-hscroll: #t)))
	    (app-text
	      (create <app-text> scrolled-frame with
		(height/char: height) (width/char: width)
		(background-color: 'white)
		(wrap: 'none)
		)))
       (pack scrolled-frame (expand: #t) (fill: 'both))
       (pack app-text)
       (send top notify-text app-text)
       app-text)]
    [() (new-app-text 24 80)]))
