;; Copyright (c) 1996 Oscar Waddell
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

(define swl:default-buffer-size 2048)

(define swl:foreign-init
  (lambda () 

    ;; The foreign-entry SWL_DoOneEvent returns
    ;;   0  no fallback to process, *args not touched
    ;;  -n  do fallback n, no args
    ;;   n  do fallback n, args have been hammered into the args array
    ;;      if the args didn't fit, then the first char of the args array
    ;;      will be #\nul and n is actually the size of string needed for
    ;;      the args
    ;; 
    (define make-event-handler
      (lambda (foreign-entry)
	(let* ((bufsize 100)
	       (buf (make-string bufsize)))
	  (rec loop
	    (lambda (block?)
              (critical-section
                (let ((result (foreign-entry block? buf bufsize)))
                  (cond
                    ((fxzero? result) (values #f #f))
                    ((fxnegative? result) (values (fx- result) #f))
                    (else 
		     (if (char=? (string-ref buf 0) #\nul)
			 (let ([newlen (fxmax result (* 2 bufsize))])
			   (set! bufsize newlen)
			   (set! buf (make-string newlen))
			   (loop block?))
			 (values result
				 (read (open-input-string buf)))))))))))))
 
    ;; Process an event on the C side.  Flag passed in determines whether
    ;; or not it should block and wait for the event.
    ;; returns 2 values:
    ;;   index of callback to invoke
    ;;   list of args to the callback (or #f if none)

    (set! swl:do-one-event
      (make-event-handler
       (foreign-procedure "SWL_DoOneEvent"
			  (boolean string integer-32) integer-32)))

    (set! swl:peek-event (foreign-procedure "SWL_PeekEvent" () boolean))

    (set! swl:raw-tcl-eval
      (lambda (str)
        (error 'swl:tcl-eval
	       "system not initialized with swl:tcl-init before ~s" str)))

    (set! swl:tcl-init
      (let ([fp-tcl-init (foreign-procedure "SWL_TclInit"
					    (integer-32 integer-32) boolean)])
	(lambda (queue-size argbuf-size)
	  (unless (fp-tcl-init queue-size argbuf-size)
	    (error 'swl:tcl-init
		   "unable to initialize graphics library (check DISPLAY)"))
	  (set! swl:raw-tcl-eval
	    (let ([foreign-tcl
		   (foreign-procedure "SWL_TclEval" (string string) string)]
		  [err-flag (make-string 1)])
	      (lambda (str)
		(let ([x (foreign-tcl str err-flag)])
		  (if (char=? (string-ref err-flag 0) #\1)
		      x
		      (error #f x)))))))))

    (set! swl:watch-channel
      ;; handler is a thunk that will be called when the given channel
      ;; is readable / writable (depends on flag passed in)
      (lambda (channel flag handler)
        (unless (and '(channel? channel) (symbol? flag))
          (error 'swl:watch-file
		 "invalid argument(s) ~s ~s" channel flag))
        (let ([cb (swl:procedure->callback handler 'swl:system-queue)])
          (swl:tcl-eval 'fileevent channel
			(case flag
			  [(read) 'readable]
			  [(write) 'writable]
			  [else (error 'swl:watch-file "invalid flag ~s" flag)])
			cb)
          cb)))

    (swl:api-procedure swl:open-tcp-client
      ;* \formdef{swl:open-tcp-client}{procedure}{(swl:open-tcp-client \var{hostname} \var{port-number} \var{async?)}}
      ;* \ret{an input-output port}
      ;* Returns an input/output port connected by a socket
      ;* to the specified port number on the given host.
      ;* If the async flag is true, the socket is connected asynchronously, i.e.,
      ;* the port object is returned without waiting for the actual connection
      ;* to be made.
      (lambda (host port async)
	(unless (and (integer? port)
		     (string? host)
		     (boolean? async))
	  (error 'swl:open-tcp-client
		 "invalid argument(s): ~s ~s ~s" port host async))
	(let ([channel (swl:tcl-eval 'socket (if async '-async #\space) host port)])
	  (swl:make-nb-input-output-port
	   (parameterize ((print-radix 10)) (format "~a:~a" host port))
	   channel swl:default-buffer-size
	   channel swl:default-buffer-size))))

    (swl:api-procedure swl:open-tcp-server
      ;* \formdef{swl:open-tcp-server}{procedure}{(swl:open-tcp-server \var{host} \var{port} \var{handler})}
      ;* \ret{an input-output port}
      ;* Creates a server socket for the given \var{port}.  \var{host} specifies
      ;* a fully qualified domain name or numerical IP address of the
      ;* network interface to use for the connection.  If \var{host} is \scheme{#f}
      ;* then the server accepts connections via any interface.
      ;* When a connection is made, \var{handler} is passed three arguments:
      ;* an input/output port connected to the client via socket, the hostname
      ;* of the client and the port number for the client.
      ;* \scheme{swl:open-tcp-server} returns a port that cannot be read from or
      ;* written to, but which can be closed to terminate the server.
      (lambda (host port handler)
        (unless (and (integer? port)
                     (or (not host) (string? host))
                     (procedure? handler))
          (error 'swl:open-tcp-server
                 "invalid argument(s): ~s ~s ~s" port host handler))
        (let ([cb (swl:procedure->callback
                   (lambda (channel hostname port-num)
                     (handler
                      (swl:make-nb-input-output-port
                       (parameterize ((print-radix 10))
                         (format "~a:~a" hostname port-num))
                       channel swl:default-buffer-size
                       channel swl:default-buffer-size)
                      (symbol->string hostname) port-num))
                   'swl:system-queue)])
          (let ([channel
                 (if host
                     (swl:tcl-eval 'socket '-server cb '-myaddr host port)
                     (swl:tcl-eval 'socket '-server cb port))])
            (swl:make-nb-input-output-port cb channel 0 channel 0)))))

    (set! swl:install-non-blocking-console
      (lambda ()
        (when (memq (machine-type) '(ppcnt i3nt))
          (error 'swl:install-non-blocking-console
                 "not supported under Windows NT"))
        (let ([p (swl:make-nb-input-output-port
                  "non-blocking console"
                  "stdin" swl:default-buffer-size
                  "stdout" swl:default-buffer-size)])
          ;;
          ;; install nonblocking port as the new global default.
          ;;
          (console-output-port 'set-global! p)
          (console-input-port 'set-global! p)
          (current-input-port 'set-global! p)
          (current-output-port 'set-global! p)
          ;;
          ;; Install nonblocking port for current thread.  This is essential
          ;; because these parameters have been previously assigned for
          ;; the thread.
          ;;
          (current-input-port p)
          (current-output-port p)
          )))

    (set! swl:get-channel
      (let ((getit (foreign-procedure "SWL_GetChannel" (string) integer-32)))
        (lambda (name)
          (getit
            (if (string? name)
                name
                (parameterize ([print-radix 10]) (format "~s" name)))))))

    (set! swl:os-get-error (foreign-procedure "SWL_GetError" () string))

    (set! swl:os-bytes-ready
      (foreign-procedure "Tcl_InputBuffered" (integer-32) integer-32))
    
    (set! swl:os-eof
      (foreign-procedure "Tcl_Eof" (integer-32) boolean))
    
    (set! swl:os-flush
      (foreign-procedure "Tcl_Flush" (integer-32) integer-32))
    
    (set! swl:os-c-read-n
      (foreign-procedure "SWL_channel_read"
                         (integer-32 string integer-32 integer-32)
                         integer-32))
    
    (set! swl:os-c-write-n
      (foreign-procedure "SWL_channel_write"
                         (integer-32 string integer-32 integer-32)
                         integer-32))

    (set! swl:name-to-window
      (foreign-procedure "SWL_name_to_window"
                         (string)
                         integer-32))

    (set! swl:map-window (foreign-procedure "Tk_MapWindow" (integer-32) void))
    (set! swl:unmap-window (foreign-procedure "Tk_UnmapWindow" (integer-32) void))

    (set! swl:move-window
      (foreign-procedure "Tk_MoveWindow"
                         (integer-32 integer-32 integer-32)
                         void))

    (set! swl:resize-window
      (foreign-procedure "Tk_ResizeWindow"
                         (integer-32 integer-32 integer-32)
                         void))

    (set! swl:debug-mem (foreign-procedure "SWL_mem_dbg" () void))

    (when (memq (machine-type) '(ppcnt i3nt))
      (load-shared-object "crtdll")
      (set! swl:open-dde-file
        (foreign-procedure "SWL_open_dde_file" (string string string)
          unsigned-32)))

; Only for old version of SWL based on Tk 4.2
;   (set! swl:query-font
;     (foreign-procedure "SWL_queryFont" (string) boolean))

    ))
